/*
 *  TOPPERS/MDCOM
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Multi Domain Communication Module
 *
 *  Copyright (C) 2016-2019 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: mdcom_kernel.c 948 2018-09-11 07:20:30Z ertl-honda $
 */

#include "mdcom.h"
#include "mdcom_config.h"
#include "target_hw_counter.h"

#define MAX_MDCOM_WAIT_EVENTS	MDCOM_NUM_FIFOCHS
#define MDCOM_TASK_SLEEP


/*
 * CONTROL OFFSET
 */
#define CTL_M2R_REQ				0x00
#define CTL_R2M_REQ				0x04
#define CTL_R2M_REQ_ARG			0x08

#define CTL_R2M_REQ_NONE		0x00
#define CTL_R2M_REQ_PROCESSEXIT	0x01
#define CTL_R2M_REQ_NOTIFY		0x02
#define CTL_R2M_REQ_PROCESSOPEN	0x03
#define CTL_R2M_REQ_REGFIFOCH	0x04
#define CTL_R2M_REQ_INITFIFOCH	0x05
#define CTL_R2M_REQ_INIT		0x87654321
#define CTL_M2R_REQ_NONE		0x00
#define CTL_M2R_REQ_INIT		0x12345678
#define CTL_M2R_REQ_ACK			0x06

#define get_r2m_req()			(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ))
#define set_r2m_req(cmd)		(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ) = cmd)
#define set_r2m_req_arg(arg)	(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ_ARG) = arg)
#define set_m2r_req(cmd)		(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_M2R_REQ) = cmd)
#define get_r2m_req_arg()		(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ_ARG))
#define get_m2r_req()			(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_M2R_REQ))

/*
 * FIFOCH M2R request
 */
#define FIFOCH_M2R_NONE   		0
#define FIFOCH_M2R_NOTIFY 		1
#define FIFOCH_M2R_INIT   		2

/*
 * FIFOCH R2M request
 */
#define FIFOCH_R2M_NONE    		0
#define FIFOCH_R2M_NOFITY  		1
#define FIFOCH_R2M_INIT    		2

/*
 *  initialized task id;
 */
static TaskType	init_task_id;

/*
 *  MDCOM Alarm state
 */
#ifdef	MDCOM_TASK_SLEEP
static bool_t	mdcom_alarm_run = false;
#endif	/* MDCOM_TASK_SLEEP */


/*
 * INTERNAL TYPES
 */
typedef struct {
	TaskType		rtos_tskid;
	uint32_t		*m2r_pending;
	uint32_t		*r2m_pending;
	uint32_t		IsAbort;
} mdcom_fifoch_event_t;


typedef struct {
	uint32_t				state;
	uint32_t				mdcom_num_process;
	mdcom_mask_t			mdcom_fifoch_mask;
	mdcom_fifoch_event_t	fifoch_events[MDCOM_NUM_FIFOCHS];
} mdcom_kernel_t;


typedef struct
{
	void		*pre_item;
	void		*next_item;
	uint32_t	timeout_downcount;
	mdcom_kernel_wait_t	*request;
} mdcom_wait_event_t;


/*
 * Variables for timeout management
 */
static mdcom_wait_event_t	*p_wait_event_1st;
static mdcom_wait_event_t	wait_events[MAX_MDCOM_WAIT_EVENTS];


/*
 * Prototypes
 */
extern void		MdcomTermHook(uint32_t cause);
static int		wait_event_with_timeout(TaskType taskid, EventMaskType target_event, mdcom_time_t timeout);
static mdcom_wait_event_t	*get_free_wait_event(void);
static int		add_wait_event(mdcom_wait_event_t *new_item);
static mdcom_wait_event_t	*remove_wait_event(mdcom_wait_event_t *remove_item);


/*
 * DEBUG
 */
//#define MDCOM_DEBUG

#define DEBUG_OUT false

#ifdef MDCOM_DEBUG

static void
__mdcom_debug_print_fifoch_event(mdcom_fifoch_event_t *ev)
{
	syslog(LOG_NOTICE, "DEBUG fifoch_event");
	syslog(LOG_NOTICE, " fifoch_event->rtos_tskid:  %d", ev->rtos_tskid_pending);
	syslog(LOG_NOTICE, " fifoch_event->m2r_pending: 0x%X (val:0x%X)", (uintptr_t)ev->m2r_pending, *(ev->m2r_pending));
	syslog(LOG_NOTICE, " fifoch_event->r2m_pending: 0x%X (val:0x%X)", (uintptr_t)ev->r2m_pending, *(ev->r2m_pending));
	syslog(LOG_NOTICE, " fifoch_event->IsAbort:     %d", ev->IsAbort);
}

#else /* !MDCOM_DEBUG */
#define  __mdcom_debug_print_fifoch_event(ev)
#endif /* MDCOM_DEBUG */

/*
 *  GCC call memcpy in case of more than 17 channes.
 */
void
*memcpy(void *dest, const void *src, size_t len)
{
	char *dst8 = (char *)dest;
	char *src8 = (char *)src;

	while (len-- > 0) {
		*dst8 = *src8;
		dst8++;
		src8++;
	}

	return dest;
}

/*
 * MODULE VARIABLES
 */
static mdcom_kernel_t the_mdcom_kernel;

static void
__mdcom_kernel_fifoch_release_waittask(const mdcom_id_t fifoch_id)
{
	mdcom_fifoch_event_t *p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];
#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	*(p_fifoch_event->r2m_pending) = FIFOCH_R2M_NONE;
	*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_NONE;
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	if (p_fifoch_event->rtos_tskid != 0) {
		p_fifoch_event->IsAbort = true;
		DEBUG(DEBUG_OUT, "Abort to wait event. taskid=%d", p_fifoch_event->rtos_tskid);
		SetEvent(p_fifoch_event->rtos_tskid, MdcomEvent);
	}
}

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
/*
 * EVENTS HANDLER (Master)
 */

/* Interrupt handler */
ISR(MdcomEvtInt)
{
	StatusType	ercd;
	int 		i;
	uint32_t 	mask;
	mdcom_fifoch_event_t *p_fifoch_event;

	// Clear Interrupt
	__mdcom_kernel_clear_int();

	DEBUG(DEBUG_OUT, "MDCOM handler");

	//Initialize
	if (get_r2m_req() == CTL_R2M_REQ_INIT) {
		DEBUG(DEBUG_OUT, "INIT");
		if (the_mdcom_kernel.state == MDCOM_KERNEL_REQINIT) {
			set_m2r_req(CTL_M2R_REQ_NONE);
			ercd = SetEvent(init_task_id, MdcomInitEvent);
			assert(ercd == E_OK);
			the_mdcom_kernel.state = MDCOM_KERNEL_INITD;
		}
	}
	//NT_OS application has been exit. Restart Multi Domain Communication Module
	else if(get_r2m_req()  == CTL_R2M_REQ_PROCESSEXIT) {
		DEBUG(DEBUG_OUT, "PROCESSEXIT");

		assert(the_mdcom_kernel.mdcom_num_process > 0);
		the_mdcom_kernel.mdcom_num_process--;
		DEBUG(DEBUG_OUT, "PROCESSEXIT. num_process = %d", the_mdcom_kernel.mdcom_num_process);

		/* update mask */
		mask = get_r2m_req_arg();
		DEBUG(DEBUG_OUT, "PROCESSEXIT. mask 0x%x", mask);
		the_mdcom_kernel.mdcom_fifoch_mask &= ~mask;
		for (i = 0; mask; i++){
			if (mask & 0x1){
				__mdcom_kernel_fifoch_release_waittask(i);
			}
			mask = mask >> 1;
		}

		DEBUG(DEBUG_OUT, "MDCOM: PROCESSEXIT");
		MdcomTermHook(TASK_EXC_ARG_PROCESSEXIT);
	}
	// signal all fifochs with pending events
	else if(get_r2m_req() == CTL_R2M_REQ_NOTIFY) {
		DEBUG(DEBUG_OUT, "NOTIFY");
		for(i=0; i < MDCOM_NUM_FIFOCHS; i++) {
			p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
			if (*(p_fifoch_event->r2m_pending) == FIFOCH_R2M_NONE) continue;
			DEBUG(DEBUG_OUT, "NOTIFY : SetEvent");
			SetEvent(p_fifoch_event->rtos_tskid, MdcomEvent);
			*(p_fifoch_event->r2m_pending) = FIFOCH_R2M_NONE;
		}
	}
	/* called from mdcom_user.c in user space */
	else if(get_r2m_req()  == CTL_R2M_REQ_PROCESSOPEN) {
		DEBUG(DEBUG_OUT, "PROCESSOPEN");
		the_mdcom_kernel.mdcom_num_process++;
		DEBUG(DEBUG_OUT, "PROCESSOPEN. num_process = %d", the_mdcom_kernel.mdcom_num_process);
	}
	else if(get_r2m_req() == CTL_R2M_REQ_REGFIFOCH) {
		DEBUG(DEBUG_OUT, "REGFIFOCH");
		mask = get_r2m_req_arg();
		DEBUG(DEBUG_OUT, "REGFIFOCH. mask 0x%x.", mask);
		the_mdcom_kernel.mdcom_fifoch_mask |= mask;
	}
	else if(get_r2m_req() == CTL_R2M_REQ_INITFIFOCH) {
		DEBUG(DEBUG_OUT, "INITFIFOCH");
		__mdcom_kernel_fifoch_init(get_r2m_req_arg());
	}
	else {
		DEBUG(DEBUG_OUT, "CTL_R2M_REQ = %u, arg = %u, state = %u",
			  get_r2m_req(),
			  get_r2m_req_arg(),
			  the_mdcom_kernel.state);
		assert_isr(0);
	}

	// Ack
	set_r2m_req(CTL_M2R_REQ_ACK);
	__mdcom_send_irq();
}

#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
/*
 * EVENTS HANDLER (Remote)
 */
ISR(MdcomEvtInt)
{
	int i;
	mdcom_fifoch_event_t *p_fifoch_event;

	// Clear Interrupt
	__mdcom_kernel_clear_int();

	DEBUG(DEBUG_OUT, "MDCOM handler");

	if (get_r2m_req() == CTL_M2R_REQ_ACK) {
		DEBUG(DEBUG_OUT, "Receive Ack.");
		the_mdcom_kernel.state = MDCOM_KERNEL_INITD;
		SetEvent(init_task_id, MdcomInitEvent);
		set_r2m_req(0);
	}

	if (the_mdcom_kernel.state != MDCOM_KERNEL_INITD) {
		return;
	}

	for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
		p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
		if (*(p_fifoch_event->m2r_pending) == FIFOCH_M2R_NONE) continue;
		if (*(p_fifoch_event->m2r_pending) != FIFOCH_M2R_NONE) {
			/* case of fifoch initialize, no queuing m2r request */
			*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_NONE;
		}
		SetEvent(p_fifoch_event->rtos_tskid, MdcomEvent);
	}
}
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

void
*memset (void *str, int c, size_t len )
{
	register char *st = (char *)str;
	while (len-- > 0) *st++ = c;
	return (void *)str;
}

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
/*
 * INIT (Master)
 */
int
__mdcom_kernel_init(mdcom_mask_t fifoch_mask, mdcom_time_t timeout, void *sregion)
{
	StatusType	ercd;
	int 		i, mdcom_ercd;
	mdcom_fifoch_event_t 	*p_fifoch_event;

	if (the_mdcom_kernel.state == MDCOM_KERNEL_INITD) {
		return MDCOM_SUCCESS;
	}

	GetTaskID(&init_task_id);

	if (the_mdcom_kernel.state == MDCOM_KERNEL_NOINIT) {
		the_mdcom_kernel.mdcom_num_process = 0;

		// Initialize wait events
		for( i = 0 ; i < MAX_MDCOM_WAIT_EVENTS ; i++ ) {
			wait_events[i].request =   NULL;
			wait_events[i].next_item = NULL;
		}
		p_wait_event_1st = NULL;

		// 1.- initialize fifoch events
		for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
			p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
			p_fifoch_event->m2r_pending = (uint32_t *)((uintptr_t)sregion + MDCOM_SREGION_PENDING_EVENTS) + i;
			p_fifoch_event->r2m_pending = p_fifoch_event->m2r_pending + MDCOM_NUM_FIFOCHS;
			__mdcom_debug_print_fifoch_event(p_fifoch_event);
		}

		the_mdcom_kernel.state = MDCOM_KERNEL_REQINIT;
		set_m2r_req(CTL_M2R_REQ_INIT);

		// Flush D-Cache
		// MASTER will invalidates the D-Cache.
		// Without flushing D-Cache, the data that has been written to the shared memory
		// will not be write back to the shared memory.
		__mdcom_dcache_flush();

		// Initialize Interrupt
		__mdcom_kernel_init_init();

		// Initialize Mutex
		__mdcom_kernel_mutex_init();

		ercd = EnableInterruptSource(MdcomEvtInt);
		assert(ercd == E_OK);

	#ifndef	MDCOM_TASK_SLEEP
		/* Activate timeout management task */
		mdcom_alarm_run = true;
		ercd = SetRelAlarm(MdcomAlarm, TICK_FOR_10MS / 10, TICK_FOR_10MS / 10);
		assert(ercd == E_OK);
	#endif	/* MDCOM_TASK_SLEEP */
	}

	mdcom_ercd = wait_event_with_timeout(init_task_id, MdcomInitEvent, timeout);
	if( (mdcom_ercd == MDCOM_SUCCESS) && (the_mdcom_kernel.state != MDCOM_KERNEL_INITD) ) {
		mdcom_ercd = MDCOM_OTHER;
	}

	return mdcom_ercd;
}
#else  /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
extern int __mdcom_kernel_regfifoch(mdcom_mask_t fifoch_mask, mdcom_time_t timeout);

/*
 * INIT
 */
int
__mdcom_kernel_init(mdcom_mask_t fifoch_mask, mdcom_time_t timeout, void *sregion)
{
	StatusType		ercd;
	EventMaskType	own_event;
	int 			i, mdcom_ercd;
	mdcom_fifoch_event_t	*p_fifoch_event;

	if (the_mdcom_kernel.state == MDCOM_KERNEL_INITD) {
		return MDCOM_SUCCESS;
	}

	if (get_m2r_req() != CTL_M2R_REQ_INIT) {
		return MDCOM_EINIT;
	}

	GetTaskID(&init_task_id);

	// Initialize wait events
	for(i = 0 ; i < MAX_MDCOM_WAIT_EVENTS ; i++) {
		wait_events[i].request =   NULL;
		wait_events[i].next_item = NULL;
	}
	p_wait_event_1st = NULL;

	/* MDCOM_IOCTL_CMD_INIT 相当の処理 */
	DEBUG(DEBUG_OUT, "Initialize channel events");
	for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
		p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
		p_fifoch_event->m2r_pending = (uint32_t *)(sregion + MDCOM_SREGION_PENDING_EVENTS) + i;
		p_fifoch_event->r2m_pending = p_fifoch_event->m2r_pending + MDCOM_NUM_FIFOCHS;
		p_fifoch_event->rtos_tskid = 0;
		__mdcom_debug_print_fifoch_event(p_fifoch_event);
	}

	the_mdcom_kernel.state = MDCOM_KERNEL_REQINIT;
	set_r2m_req(CTL_R2M_REQ_INIT);

	__mdcom_dcache_flush();

	// Initialize Interrupt
	__mdcom_kernel_init_init();

	__mdcom_send_irq();

	ercd = EnableInterruptSource(MdcomEvtInt);
	assert(ercd == E_OK);

	mdcom_ercd = wait_event_with_timeout(init_task_id, MdcomInitEvent, timeout);
	if (mdcom_ercd != MDCOM_SUCCESS) {
		return mdcom_ercd;
	}

	/* MDCOM_IOCTL_CMD_REGFIFOCH 相当の処理 */
	mdcom_ercd = __mdcom_kernel_regfifoch(fifoch_mask, timeout);

	return mdcom_ercd;
}

int
__mdcom_kernel_regfifoch(mdcom_mask_t fifoch_mask, mdcom_time_t timeout)
{
	StatusType			ercd;
	TaskType			own_task_id;
	EventMaskType		own_event;
	mdcom_kernel_wait_t wait_request;
	int					mdcom_ercd;

	if (!fifoch_mask){
		return MDCOM_SUCCESS;
	}

	/* set mask value */
	the_mdcom_kernel.mdcom_fifoch_mask = fifoch_mask;

	set_r2m_req(CTL_R2M_REQ_REGFIFOCH);
	set_r2m_req_arg(fifoch_mask);
	__mdcom_send_irq();

	GetTaskID(&own_task_id);

	mdcom_ercd = wait_event_with_timeout(own_task_id, MdcomInitEvent, timeout);

	return mdcom_ercd;
}
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

/*
 * TERM
 */
int
__mdcom_kernel_term(void)
{
	StatusType	ercd;
	int	 		i;

	ercd = DisableInterruptSource(MdcomEvtInt);
	assert(ercd == E_OK);

	the_mdcom_kernel.state = MDCOM_KERNEL_NOINIT;

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	__mdcom_kernel_mutex_init();
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

	for(i=0; i < MDCOM_NUM_FIFOCHS; i++) {
		__mdcom_kernel_fifoch_release_waittask(i);
	}

	DEBUG(DEBUG_OUT, "MDCOM: TASK_EXC_ARG_TERM.");
	MdcomTermHook(TASK_EXC_ARG_TERM);

	return MDCOM_SUCCESS;
}

/*
 *  Get Driver State
 */
int
__mdcom_kernel_get_state(void)
{
	return the_mdcom_kernel.state;
}

/*
 * INIT Shared Region(Memory)
 */
int
__mdcom_kernel_sregion_init(void **sregion)
{
	*sregion = (void *)SREGION_PHYS_ADDRESS;
#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	memset(*sregion, 0, MDCOM_SREGION_SIZE);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

	return MDCOM_SUCCESS;
}

int
__mdcom_kernel_notify(mdcom_id_t fifoch_id)
{
	mdcom_fifoch_event_t *p_fifoch_event;

	p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];
#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_NOTIFY;
#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
	*(p_fifoch_event->r2m_pending) = FIFOCH_R2M_NOFITY;
	set_r2m_req(CTL_R2M_REQ_NOTIFY);
	set_r2m_req_arg(fifoch_id);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	
	__mdcom_send_irq();

	return MDCOM_SUCCESS;
}

int
__mdcom_kernel_wait(mdcom_id_t fifoch_id, mdcom_time_t timeout)
{
	TaskType	own_task_id;
	int 		mdcom_ercd;
	mdcom_fifoch_event_t	*p_fifoch_event;

	p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];

	__mdcom_debug_print_fifoch_event(p_fifoch_event);

	if (p_fifoch_event->rtos_tskid != 0) return MDCOM_WAITOVER;

	GetTaskID(&own_task_id);
	p_fifoch_event->rtos_tskid = own_task_id;
	p_fifoch_event->IsAbort =    false;

	mdcom_ercd = wait_event_with_timeout(own_task_id, MdcomEvent, timeout);
	p_fifoch_event->rtos_tskid = 0;
	if (p_fifoch_event->IsAbort == true) {
		mdcom_ercd = MDCOM_OTHER;
	}

	return mdcom_ercd;
}

int
__mdcom_kernel_get_fifoch_mask(mdcom_mask_t *p_fifoch_mask)
{
	*p_fifoch_mask = the_mdcom_kernel.mdcom_fifoch_mask;
	return MDCOM_SUCCESS;
}

/*
 * Timeout management
 */
static int
wait_event_with_timeout(TaskType taskid, EventMaskType target_event, mdcom_time_t timeout)
{
	StatusType			ercd;
	EventMaskType		own_event;
	mdcom_kernel_wait_t	wait_request;

	if (timeout == -1) {
		/* 無限待ち */
		ercd = WaitEvent(target_event);
		if( ercd == E_OK )  {
			ClearEvent(target_event);
		}
	}
	else if(timeout == 0) {
		/* ポーリング時 */
		ercd = GetEvent(taskid, &own_event);
		if (ercd == E_OK) {
			if ((own_event & target_event) == target_event)  {
				ClearEvent(target_event);
			}
			else {
				return MDCOM_TIMEOUT;
			}
		}
	}
	else
	{
		/* タイムアウト時間設定時 */
		wait_request.rtos_tskid =   taskid;
		wait_request.rtos_event =   target_event;
		wait_request.timeout_time = timeout;
		wait_request.IsTimeouted =  false;
		if( __mdcom_kernel_set_timeout(&wait_request) != 0 ) {
			return MDCOM_OTHER;
		}
		ercd = WaitEvent(target_event);
		if (ercd == E_OK) {
			ClearEvent(target_event);
			if (wait_request.IsTimeouted == true) {
				return MDCOM_TIMEOUT;
			}
			else {
				__mdcom_kernel_clear_timeout(&wait_request);
			}
		}
	}

	if (ercd != E_OK) {
		return MDCOM_OTHER;
	}

	return MDCOM_SUCCESS;
}


int
__mdcom_kernel_set_timeout(mdcom_kernel_wait_t *wait_request)
{
	mdcom_wait_event_t	*new_item;
	StatusType	ercd;
	ISRType		isrid;
	int			retcd = -1;
	int			is_1st_event;

	/* Make sure current context isn't ISR. */
	isrid = GetISRID();
	if (isrid != INVALID_ISR) {
		return -1;
	}

	__mdcom_kernel_lock();
	/* Get free area of wait_event store area */
	new_item = get_free_wait_event();
	if (new_item != NULL) {
		/* Create new item */
		new_item->timeout_downcount = wait_request->timeout_time;
		new_item->request =   wait_request;
		/* Add new item to wait event queue */
		is_1st_event = add_wait_event(new_item);
		__mdcom_kernel_unlock();
	#ifdef	MDCOM_TASK_SLEEP
		if ((is_1st_event == true) && (mdcom_alarm_run == false)) {
			/* Activate timeout management task */
			mdcom_alarm_run = true;
			ercd = SetRelAlarm(MdcomAlarm, TICK_FOR_10MS / 10, TICK_FOR_10MS / 10);
			assert(ercd == E_OK);
		}
	#endif	/* MDCOM_TASK_SLEEP */
		retcd = 0;
	}
	else {
		__mdcom_kernel_unlock();
		syslog(LOG_ERROR, "MDCOM : ERROR - Too many wait events.");
	}

	return retcd;
}


int
__mdcom_kernel_clear_timeout(mdcom_kernel_wait_t *wait_request)
{
	mdcom_wait_event_t	*crnt_item;
	ISRType		isrid;

	/* Make sure current context isn't ISR. */
	isrid = GetISRID();
	if (isrid != INVALID_ISR) {
		return -1;
	}

	__mdcom_kernel_lock();
	/* find target wait event */
	crnt_item = p_wait_event_1st;
	while(crnt_item != NULL) {
		if (crnt_item->request == wait_request) {
			/* Remove wait event */
			remove_wait_event(crnt_item);
			break;
		}
		crnt_item = crnt_item->next_item;
	}
	__mdcom_kernel_unlock();

	return 0;
}


TASK(MdcomTask)
{
	mdcom_wait_event_t	*crnt_item;
	mdcom_kernel_wait_t *wait_request;
	StatusType	ercd;

	crnt_item = p_wait_event_1st;
	while (crnt_item != NULL) {
		(crnt_item->timeout_downcount)--;
		if (crnt_item->timeout_downcount == 0) {
			wait_request = crnt_item->request;
			wait_request->IsTimeouted = true;
			crnt_item = remove_wait_event(crnt_item);
			ercd = SetEvent(wait_request->rtos_tskid, wait_request->rtos_event);
			assert(ercd == E_OK);
		}
		else {
			crnt_item = crnt_item->next_item;
		}
	}

#ifdef	MDCOM_TASK_SLEEP
	if ((p_wait_event_1st == NULL) && (mdcom_alarm_run == true)) {
		CancelAlarm(MdcomAlarm);
		mdcom_alarm_run = false;
	}
#endif	/* MDCOM_TASK_SLEEP */

	TerminateTask();
}


static mdcom_wait_event_t
*get_free_wait_event(void)
{
	mdcom_wait_event_t	*free_item = NULL;
	int		i;

	for (i = 0; i < MAX_MDCOM_WAIT_EVENTS; i++) {
		if (wait_events[i].request == NULL) {
			free_item = &wait_events[i];
			break;
		}
	}

	return free_item;
}


static int
add_wait_event(mdcom_wait_event_t *new_item)
{
	mdcom_wait_event_t	*pre_item, *crnt_item, *next_item;

	if (p_wait_event_1st != NULL) {
		crnt_item = p_wait_event_1st;
		while(1) {
			pre_item =  crnt_item->pre_item;
			next_item = crnt_item->next_item;

			if (new_item->timeout_downcount < crnt_item->timeout_downcount) {
				/* Add new item to first or previous */
				crnt_item->pre_item =  new_item;
				new_item->pre_item =   pre_item;
				new_item->next_item =  crnt_item;
				if (pre_item != NULL) {
					/* Add to previous */
					pre_item->next_item =  new_item;
					return false;
				}
				else {
					/* Add to first */
					p_wait_event_1st = new_item;
					return true;
				}
			}

			if (next_item != NULL) {
				crnt_item = next_item;
			}
			else {
				/* Add new item to bottom */
				crnt_item->next_item = new_item;
				new_item->pre_item =  crnt_item;
				new_item->next_item = NULL;

				return false;
			}
		}
	}
	else {
		/* Add new item to first */
		new_item->pre_item =  NULL;
		new_item->next_item = NULL;
		p_wait_event_1st =    new_item;

		return true;
	}
}


static mdcom_wait_event_t
*remove_wait_event(mdcom_wait_event_t *remove_item)
{
	mdcom_wait_event_t	*pre_item, *next_item;

	pre_item =  remove_item->pre_item;
	next_item = remove_item->next_item;
	if (pre_item != NULL) {
		pre_item->next_item =  next_item;
		if (next_item != NULL) {
			next_item->pre_item =  pre_item;
		}
	}
	else {
		p_wait_event_1st = next_item;
		if (next_item != NULL) {
			next_item->pre_item =  NULL;
		}
	}

	remove_item->request = NULL;

	return next_item;
}


/*
 * Function in mdcom_user.c
 */
extern void __mdcom_fifoch_init(mdcom_id_t fifoch_id);

int
__mdcom_kernel_fifoch_init(const mdcom_id_t fifoch_id)
{
	mdcom_fifoch_event_t *p_fifoch_event;
	p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];

	DEBUG(DEBUG_OUT, "FIFO INIT", p_fifoch_event->rtos_tskid);

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	__mdcom_kernel_lock();
	__mdcom_fifoch_init(fifoch_id);
	__mdcom_kernel_unlock();

	__mdcom_kernel_fifoch_release_waittask(fifoch_id);

	*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_INIT;
#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
	set_r2m_req(CTL_R2M_REQ_INITFIFOCH);
	set_r2m_req_arg(fifoch_id);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	__mdcom_send_irq();

	return MDCOM_SUCCESS;
}

int
__mdcom_kernel_get_num_process(uint32_t *p_num_process)
{
	*p_num_process = the_mdcom_kernel.mdcom_num_process;
	return MDCOM_SUCCESS;
}

void
__mdcom_kernel_lock()
{
#ifndef TNUM_SPINLOCK
	/* シングルコア時 */
	SuspendOSInterrupts();
#else /* TNUM_SPINLOCK */
	/* マルチコア時 */
	GetSpinlock(MdcomSpinLock);
#endif /* TNUM_SPINLOCK */
}

void
__mdcom_kernel_unlock()
{
#ifndef TNUM_SPINLOCK
	/* シングルコア時 */
	ResumeOSInterrupts();
#else /* TNUM_SPINLOCK */
	/* マルチコア時 */
	ReleaseSpinlock(MdcomSpinLock);
#endif /* TNUM_SPINLOCK */
}
