/*
 *  TOPPERS/MDCOM
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Multi Domain Communication Module
 *
 *  Copyright (C) 2016-2018 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: mdcom_kernel.c 948 2018-09-11 07:20:30Z ertl-honda $
 */

#include "mdcom.h"
#include "mdcom_config.h"
#include "kernel_cfg.h"

/*
 * CONTROL OFFSET
 */
#define CTL_M2R_REQ		0x00
#define CTL_R2M_REQ		0x04
#define CTL_R2M_REQ_ARG	0x08

#define CTL_R2M_REQ_NONE			0x00
#define CTL_R2M_REQ_PROCESSEXIT	0x01
#define CTL_R2M_REQ_NOTIFY			0x02
#define CTL_R2M_REQ_PROCESSOPEN	0x03
#define CTL_R2M_REQ_REGFIFOCH		0x04
#define CTL_R2M_REQ_INITFIFOCH		0x05
#define CTL_R2M_REQ_INIT			0x87654321
#define CTL_M2R_REQ_NONE			0x00
#define CTL_M2R_REQ_INIT			0x12345678
#define CTL_M2R_REQ_ACK			0x06

#define get_r2m_req()			(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ))
#define set_r2m_req(cmd)		(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ) = cmd)
#define set_r2m_req_arg(arg)	(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ_ARG) = arg)
#define set_m2r_req(cmd)		(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_M2R_REQ) = cmd)
#define get_r2m_req_arg()		(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_R2M_REQ_ARG))
#define get_m2r_req()			(*(uint32_t *)(SREGION_PHYS_ADDRESS + CTL_M2R_REQ))

/*
 * FIFOCH M2R request
 */
#define FIFOCH_M2R_NONE   0
#define FIFOCH_M2R_NOTIFY 1
#define FIFOCH_M2R_INIT   2

/*
 * FIFOCH R2M request
 */
#define FIFOCH_R2M_NONE    0
#define FIFOCH_R2M_NOFITY  1
#define FIFOCH_R2M_INIT    2

/*
 *  initialized task id;
 */
static ID init_task_id;

/*
 * INTERNAL TYPES
 */
typedef struct {
	ID			rtos_sem;
	ID			rtos_tskid;
	uint32_t	*m2r_pending;
	uint32_t	*r2m_pending;
} mdcom_fifoch_event_t;


typedef struct {
	uint32_t				state;
	uint32_t				mdcom_num_process;
	mdcom_mask_t			mdcom_fifoch_mask;
	mdcom_fifoch_event_t	fifoch_events[MDCOM_NUM_FIFOCHS];
} mdcom_kernel_t;

/*
 * DEBUG
 */
//#define MDCOM_DEBUG

#define DEBUG_OUT false

#ifdef MDCOM_DEBUG

static void
__mdcom_debug_print_fifoch_event(mdcom_fifoch_event_t *ev)
{
	ER      ercd;
	T_RSEM  rsem;

	syslog(LOG_NOTICE, "DEBUG fifoch_event");
	syslog(LOG_NOTICE, " fifoch_event->m2r_pending: 0x%X (val:0x%X)", (uint32_t)ev->m2r_pending, *(ev->m2r_pending));
	syslog(LOG_NOTICE, " fifoch_event->r2m_pending: 0x%X (val:0x%X)", (uint32_t)ev->r2m_pending, *(ev->r2m_pending));

	ercd = ref_sem(ev->rtos_sem, &rsem);
	assert(ercd == E_OK);

	if (rsem.wtskid == TSK_NONE) {
		syslog(LOG_NOTICE, " rtos_sem no task waiting");
	} else {
		syslog(LOG_NOTICE, " rtos_sem task with id %d waiting", rsem.wtskid);
	}
	syslog(LOG_NOTICE, " rtos_sem value: %d", rsem.semcnt);
}

#else /* !MDCOM_DEBUG */
#define  __mdcom_debug_print_fifoch_event(ev)
#endif /* MDCOM_DEBUG */

/*
 *  GCC call memcpy in case of more than 17 channes.
 */
void
*memcpy(void *dest, const void *src, size_t len)
{
	while (len-- > 0)
    {
		*((char *)dest) = *((char *)src);
		dest++;
		src++;
	}

	return( dest );
}

/*
 * MODULE VARIABLES
 */
static mdcom_kernel_t the_mdcom_kernel;

static void
__mdcom_kernel_fifoch_release_waittask(const mdcom_id_t fifoch_id)
{
	mdcom_fifoch_event_t *p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];
#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	*(p_fifoch_event->r2m_pending) = FIFOCH_R2M_NONE;
	*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_NONE;
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	if (p_fifoch_event->rtos_tskid != 0){
		DEBUG(DEBUG_OUT, "irel_wai to task %d", p_fifoch_event->rtos_tskid);
		if(sns_ctx()) {
			irel_wai(p_fifoch_event->rtos_tskid);
		}
		else {
			rel_wai(p_fifoch_event->rtos_tskid);
		}
	}
}

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
/*
 * EVENTS HANDLER (Master)
 */
void mdcom_handler(intptr_t exinf)
{
	ER ercd;
	int i;
	uint32_t mask;
	mdcom_fifoch_event_t *p_fifoch_event;

	// Clear Interrupt
	__mdcom_kernel_clear_int();

	DEBUG(DEBUG_OUT, "MDCOM handler");

	//Initialize
	if (get_r2m_req() == CTL_R2M_REQ_INIT){
		DEBUG(DEBUG_OUT, "INIT");
		if (the_mdcom_kernel.state == MDCOM_KERNEL_REQINIT) {
			set_m2r_req(CTL_M2R_REQ_NONE);
			ercd = isig_sem(MDCOM_SEM_INIT);
			assert(ercd == E_OK);
			the_mdcom_kernel.state = MDCOM_KERNEL_INITD;
		}
	}
	//NT_OS application has been exit. Restart Multi Domain Communication Module
	else if(get_r2m_req()  == CTL_R2M_REQ_PROCESSEXIT) {
		DEBUG(DEBUG_OUT, "PROCESSEXIT");

		assert(the_mdcom_kernel.mdcom_num_process > 0);
		the_mdcom_kernel.mdcom_num_process--;
		DEBUG(DEBUG_OUT, "PROCESSEXIT. num_process = %d", the_mdcom_kernel.mdcom_num_process);

		/* update mask */
		mask = get_r2m_req_arg();
		DEBUG(DEBUG_OUT, "PROCESSEXIT. mask 0x%x", mask);
		the_mdcom_kernel.mdcom_fifoch_mask &= ~mask;
		for (i = 0; mask; i++){
			if (mask & 0x1){
				__mdcom_kernel_fifoch_release_waittask(i);
			}
			mask = mask >> 1;
		}
		DEBUG(DEBUG_OUT, "PROCESSEXIT. iras_tex(%d)", init_task_id);
		iras_tex(init_task_id, TASK_EXC_ARG_PROCESSEXIT);
	}
	// signal all fifochs with pending events
	else if(get_r2m_req() == CTL_R2M_REQ_NOTIFY) {
		DEBUG(DEBUG_OUT, "NOTIFY");
		for(i=0; i < MDCOM_NUM_FIFOCHS; i++) {
			p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
			if (*(p_fifoch_event->r2m_pending) == FIFOCH_R2M_NONE) continue;
			DEBUG(DEBUG_OUT, "NOTIFY : isig_sem");
			isig_sem(p_fifoch_event->rtos_sem);
			*(p_fifoch_event->r2m_pending) = FIFOCH_R2M_NONE;
		}
	}
	/* called from mdcom_user.c in user space */
	else if(get_r2m_req()  == CTL_R2M_REQ_PROCESSOPEN) {
		DEBUG(DEBUG_OUT, "PROCESSOPEN");
		the_mdcom_kernel.mdcom_num_process++;
		DEBUG(DEBUG_OUT, "PROCESSOPEN. num_process = %d", the_mdcom_kernel.mdcom_num_process);
	}
	else if(get_r2m_req() == CTL_R2M_REQ_REGFIFOCH) {
		DEBUG(DEBUG_OUT, "REGFIFOCH");
		mask = get_r2m_req_arg();
		DEBUG(DEBUG_OUT, "REGFIFOCH. mask 0x%x.", mask);
		the_mdcom_kernel.mdcom_fifoch_mask |= mask;
	}
	else if(get_r2m_req() == CTL_R2M_REQ_INITFIFOCH) {
		DEBUG(DEBUG_OUT, "INITFIFOCH");
		__mdcom_kernel_fifoch_init(get_r2m_req_arg());
	}
	else {
		DEBUG(DEBUG_OUT, "CTL_R2M_REQ = %u, arg = %u, state = %u",
			  get_r2m_req(),
			  get_r2m_req_arg(),
			  the_mdcom_kernel.state);
		assert(0);
	}

	// Ack
	set_r2m_req(CTL_M2R_REQ_ACK);
	__mdcom_send_irq();
}

#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
/*
 * EVENTS HANDLER (Remote)
 */
void mdcom_handler(intptr_t exinf)
{
	int i;
	mdcom_fifoch_event_t *p_fifoch_event;

	// Clear Interrupt
	__mdcom_kernel_clear_int();

	DEBUG(DEBUG_OUT, "MDCOM handler");

	if (get_r2m_req() == CTL_M2R_REQ_ACK) {
		DEBUG(DEBUG_OUT, "Receive Ack.");
		the_mdcom_kernel.state = MDCOM_KERNEL_INITD;
		isig_sem(MDCOM_SEM_INIT);
		set_r2m_req(0);
	}

	if (the_mdcom_kernel.state != MDCOM_KERNEL_INITD) {
		return;
	}

	for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
		p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
		if (*(p_fifoch_event->m2r_pending) == FIFOCH_M2R_NONE) continue;
		if (*(p_fifoch_event->m2r_pending) != FIFOCH_M2R_NONE) {
			/* case of fifoch initialize, no queuing m2r request */
			*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_NONE;
		}
		isig_sem(p_fifoch_event->rtos_sem);
	}
}
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

void
*memset (void *str, int c, size_t len )
{
	register char *st = (char *)str;
	while (len-- > 0) *st++ = c;
	return (void *)str;
}

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
/*
 * INIT (Master)
 */
int
__mdcom_kernel_init(mdcom_mask_t fifoch_mask, mdcom_time_t timeout, void *sregion)
{
	ER ret;
	int i;
	mdcom_fifoch_event_t *p_fifoch_event;
	ID semaphores[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_SEMAPHORES;

	if (the_mdcom_kernel.state == MDCOM_KERNEL_INITD) {
		return 0;
	}

	if (the_mdcom_kernel.state == MDCOM_KERNEL_NOINIT) {
		the_mdcom_kernel.mdcom_num_process = 0;

		get_tid(&init_task_id);

		// 1.- initialize fifoch events
		for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
			p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
			p_fifoch_event->rtos_sem = semaphores[i];
			p_fifoch_event->m2r_pending = (uint32_t *)(sregion + MDCOM_SREGION_PENDING_EVENTS) + i;
			p_fifoch_event->r2m_pending = p_fifoch_event->m2r_pending + MDCOM_NUM_FIFOCHS;
			__mdcom_debug_print_fifoch_event(p_fifoch_event);
		}

		the_mdcom_kernel.state = MDCOM_KERNEL_REQINIT;
		set_m2r_req(CTL_M2R_REQ_INIT);

		// Flush D-Cache
		// MASTER will invalidates the D-Cache.
		// Without flushing D-Cache, the data that has been written to the shared memory
		// will not be write back to the shared memory.
		__mdcom_dcache_flush();

		// Initialize Interrupt
		__mdcom_kernel_init_init();

		// Initialize Mutex
		__mdcom_kernel_mutex_init();

		ret = ena_int(MDCOM_INTNO);
		assert(ret == E_OK);
	}

	ret = twai_sem(MDCOM_SEM_INIT, timeout);

	if (the_mdcom_kernel.state == MDCOM_KERNEL_INITD) return 0;
	if (ret == E_TMOUT) return MDCOM_TIMEOUT;

	return MDCOM_OTHER;
}
#else  /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
extern int __mdcom_kernel_regfifoch(mdcom_mask_t fifoch_mask, mdcom_time_t timeout);

/*
 * INIT
 */
int
__mdcom_kernel_init(mdcom_mask_t fifoch_mask, mdcom_time_t timeout, void *sregion)
{
	ER ret;
	int i;
	mdcom_fifoch_event_t *p_fifoch_event;
	ID semaphores[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_SEMAPHORES;

	if (the_mdcom_kernel.state == MDCOM_KERNEL_INITD) {
		return 0;
	}

	if (get_m2r_req() != CTL_M2R_REQ_INIT) {
		return MDCOM_EINIT;
	}

	get_tid(&init_task_id);

	/* MDCOM_IOCTL_CMD_INIT 相当の処理 */
	DEBUG(DEBUG_OUT, "Initialize channel events");
	for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
		p_fifoch_event = &the_mdcom_kernel.fifoch_events[i];
		p_fifoch_event->rtos_sem = semaphores[i];
		p_fifoch_event->m2r_pending = (uint32_t *)(sregion + MDCOM_SREGION_PENDING_EVENTS) + i;
		p_fifoch_event->r2m_pending = p_fifoch_event->m2r_pending + MDCOM_NUM_FIFOCHS;
		p_fifoch_event->rtos_tskid = 0;
		__mdcom_debug_print_fifoch_event(p_fifoch_event);
	}

	the_mdcom_kernel.state = MDCOM_KERNEL_REQINIT;
	set_r2m_req(CTL_R2M_REQ_INIT);

	__mdcom_dcache_flush();

	// Initialize Interrupt
	__mdcom_kernel_init_init();

	__mdcom_send_irq();

	ret = ena_int(MDCOM_INTNO);
	assert(ret == E_OK);

	ret = twai_sem(MDCOM_SEM_INIT, timeout);
	if (ret == E_TMOUT) return MDCOM_TIMEOUT;
	if (ret != E_OK) return MDCOM_OTHER;

	/* MDCOM_IOCTL_CMD_REGFIFOCH 相当の処理 */
	ret = __mdcom_kernel_regfifoch(fifoch_mask, timeout);
	if (ret != 0)	return ret;
	
	return 0;
}

int
__mdcom_kernel_regfifoch(mdcom_mask_t fifoch_mask, mdcom_time_t timeout)
{
	ER ret;

	if (!fifoch_mask){
		return 0;
	}

	/* set mask value */
	the_mdcom_kernel.mdcom_fifoch_mask = fifoch_mask;

	set_r2m_req(CTL_R2M_REQ_REGFIFOCH);
	set_r2m_req_arg(fifoch_mask);
	__mdcom_send_irq();

	/* とりあえず timeout 指定可能に */
	ret = twai_sem(MDCOM_SEM_INIT, timeout);
	if (ret == E_TMOUT) return MDCOM_TIMEOUT;
	if (ret != E_OK) return MDCOM_OTHER;

	return 0;
}
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

/*
 * TERM
 */
int
__mdcom_kernel_term(void)
{
	int i;
	ER ret;

	ret = dis_int(MDCOM_INTNO);
	assert(ret == E_OK);

	the_mdcom_kernel.state = MDCOM_KERNEL_NOINIT;

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	__mdcom_kernel_mutex_init();
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

	for(i=0; i < MDCOM_NUM_FIFOCHS; i++) {
		__mdcom_kernel_fifoch_release_waittask(i);
	}

	ras_tex(init_task_id, TASK_EXC_ARG_TERM);

	return 0;
}

/*
 *  Get Driver State
 */
int
__mdcom_kernel_get_state(void)
{
	return the_mdcom_kernel.state;
}

/*
 * INIT Shared Region(Memory)
 */
int
__mdcom_kernel_sregion_init(void **sregion)
{
	*sregion = (void *)SREGION_PHYS_ADDRESS;
#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	memset(*sregion, 0, MDCOM_SREGION_SIZE);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	return 0;
}

int
__mdcom_kernel_notify(mdcom_id_t fifoch_id)
{
	mdcom_fifoch_event_t *p_fifoch_event;

	p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];
#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_NOTIFY;
#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
	*(p_fifoch_event->r2m_pending) = FIFOCH_R2M_NOFITY;
	set_r2m_req(CTL_R2M_REQ_NOTIFY);
	set_r2m_req_arg(fifoch_id);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	
	__mdcom_send_irq();

	return 0;
}

int
__mdcom_kernel_wait(mdcom_id_t fifoch_id, mdcom_time_t timeout)
{
	mdcom_fifoch_event_t *p_fifoch_event;
	ER ercd;
	ID tskid;

	p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];

	__mdcom_debug_print_fifoch_event(p_fifoch_event);

	if (p_fifoch_event->rtos_tskid != 0) return MDCOM_WAITOVER;

	get_tid(&tskid);
	p_fifoch_event->rtos_tskid = tskid;
	ercd = twai_sem(p_fifoch_event->rtos_sem, (TMO)timeout);
	p_fifoch_event->rtos_tskid = 0;

	if (ercd == E_TMOUT) return MDCOM_TIMEOUT;
	if (ercd == E_RLWAI) return MDCOM_RESTART;
	if (ercd != E_OK) return MDCOM_OTHER;

	return 0;
}

int
__mdcom_kernel_get_fifoch_mask(mdcom_mask_t *p_fifoch_mask)
{
	*p_fifoch_mask = the_mdcom_kernel.mdcom_fifoch_mask;
	return MDCOM_SUCCESS;
}

/*
 * Function in mdcom_user.c
 */
extern void __mdcom_fifoch_init(mdcom_id_t fifoch_id);

int
__mdcom_kernel_fifoch_init(const mdcom_id_t fifoch_id)
{
	mdcom_fifoch_event_t *p_fifoch_event;
	p_fifoch_event = &the_mdcom_kernel.fifoch_events[fifoch_id];

	DEBUG(DEBUG_OUT, "FIFO INIT", p_fifoch_event->rtos_tskid);

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	__mdcom_kernel_lock();
	__mdcom_fifoch_init(fifoch_id);
	__mdcom_kernel_unlock();

	__mdcom_kernel_fifoch_release_waittask(fifoch_id);

	*(p_fifoch_event->m2r_pending) = FIFOCH_M2R_INIT;
#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */
	set_r2m_req(CTL_R2M_REQ_INITFIFOCH);
	set_r2m_req_arg(fifoch_id);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	__mdcom_send_irq();

	return MDCOM_SUCCESS;
}

int
__mdcom_kernel_get_num_process(uint32_t *p_num_process)
{
	*p_num_process = the_mdcom_kernel.mdcom_num_process;
	return MDCOM_SUCCESS;
}

#if TKERNEL_PRID == 0x0007
/*
 *  For ASP Kernel
 */
void
__mdcom_kernel_lock()
{
	if(sns_ctx()) {
		iloc_cpu();
	}
	else {
		loc_cpu();
	}
}

void
__mdcom_kernel_unlock()
{
	if(sns_ctx()) {
		iunl_cpu();
	}
	else {
		unl_cpu();
	}
}

#endif /* TKERNEL_PRID == 0x0007 */

#if TKERNEL_PRID == 0x0008
/*
 *  For FMP Kernel
 */
void
__mdcom_kernel_lock()
{
	if(sns_ctx()) {
		iloc_spn(MDCOM_SPN);
	}
	else {
		loc_spn(MDCOM_SPN);
	}
}

void
__mdcom_kernel_unlock()
{
	if(sns_ctx()) {
		iunl_spn(MDCOM_SPN);
	}
	else {
		unl_spn(MDCOM_SPN);
	}
}
#endif /* TKERNEL_PRID == 0x0008 */
